/* qcow2raw.c
 *
 * Generates raw image data from an existing qcow image
 *
 * (c) 2006 Julian Chesterfield and Andrew Warfield
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation; or, when distributed
 * separately from the Linux kernel or incorporated into other
 * software packages, subject to the following license:
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this source file (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify,
 * merge, publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/statvfs.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <string.h>
#include "tapdisk.h"

#define TAPDISK 1
#define BLOCK_PROCESSSZ 4096

static int maxfds, *qcowio_fd, aio_fd, running = 1, complete = 0; 
static int returned_read_events = 0, returned_write_events = 0;
static int submit_events = 0;
static uint32_t read_idx = 0, write_idx = 0;
struct disk_driver ddqcow;
static uint64_t prev = 0, written = 0;
static char output[25];

void print_bytes(void *ptr, int length) {

  int i,k;
  unsigned char *p = ptr;

    printf("Buf dump, length %d:\n",length);
    for (k = 0; k < length; k++) {
        printf("%x",*p);
        *p++;
	if (k % 16 == 0) printf("\n");
        else if (k % 2 == 0) printf(" ");	
    }
    printf("\n");
    return;
}

void debug_output(uint64_t progress, uint64_t size)
{
	/*Output progress every 5% */	
	uint64_t blocks = size/20;

	if (progress/blocks > prev) {
		memcpy(output+prev+1,"=>",2);
		prev++;
		printf("\r%s     %llu%%", 
			output, (long long)((prev-1)*5));
		fflush(stdout);
	}
	return;
}

static inline void LOCAL_FD_SET(fd_set *readfds) 
{
	FD_SET(qcowio_fd[0], readfds);
	
	maxfds = qcowio_fd[0] + 1;
	return;
}

static int send_read_responses(struct disk_driver *dd, int res, uint64_t sec,
			       int nr_secs, int idx, void *private)
{
	int ret;

	if (res < 0) {
		 if (res != -EBUSY)
			printf("AIO FAILURE: res [%d]!\n",res);
		 return res;
	}

	returned_read_events++;
	read_idx = idx;
	
	ret = pwrite(aio_fd,private,BLOCK_PROCESSSZ, (off_t)sec << 9);
	if (ret == -1) {
		printf("AIO FAILURE: can't write data\n");
		exit(-1);
	}
	written += ret;
	returned_write_events++;

	debug_output(written, dd->td_state->size << 9);
	return 0;
}

int main(int argc, char *argv[])
{
	int ret = -1, fd, len,input;
	long int size;
	fd_set readfds;
	struct timeval timeout = { 0, 0 };
	uint64_t i;
	char *buf;
	struct stat finfo;

	if (argc != 3) {
		fprintf(stderr, "Qcow-utils: v1.0.0\n");
		fprintf(stderr, "usage: %s <Dest File descriptor> "
			"<Qcow SRC IMAGE>\n", 
		       argv[0]);
		exit(-1);
	}

	ddqcow.td_state = malloc(sizeof(struct td_state));
	
	/*Open qcow source file*/	
	ddqcow.drv = &tapdisk_qcow;
	ddqcow.private = malloc(ddqcow.drv->private_data_size);

        if (ddqcow.drv->td_open(&ddqcow, argv[2], TD_RDONLY)!=0) {
		printf("Unable to open Qcow file [%s]\n",argv[2]);
		exit(-1);
	} else printf("QCOW file opened, size %llu\n",
		      (long long unsigned)ddqcow.td_state->size);

	qcowio_fd = ddqcow.io_fd;

        /*Setup aio destination file*/
	ret = stat(argv[1],&finfo);
	if (ret == -1) {
		/*Check errno*/
		switch(errno) {
		case ENOENT:
			/*File doesn't exist, create*/
			fd = open(argv[1], 
				  O_RDWR | O_LARGEFILE | O_CREAT, 0644);
			if (fd < 0) {
				printf("ERROR creating file [%s] "
					 "(errno %d)\n",
				       argv[1], 0 - errno);
				exit(-1);
			}
			if (ftruncate(fd, (off_t)ddqcow.td_state->size<<9) < 0) {
				printf("Unable to create file "
					"[%s] of size %llu (errno %d). "
					 "Exiting...\n",
					argv[1], 
					(long long unsigned)ddqcow.td_state->size<<9, 
					0 - errno);
				close(fd);
				exit(-1);
			}
			close(fd);
			break;
		case  ENXIO:
			printf("ERROR Device [%s] does not exist\n",argv[1]);
			exit(-1);
		default: 
			printf("An error occurred opening Device [%s] "
				 "(errno %d)\n",
			       argv[1], 0 - errno);
			exit(-1);
		}
	} else {		
		fprintf(stderr, "WARNING: All existing data in "
			"%s will be overwritten.\nDo you wish to continue? "
			"(y or n)  ",
			argv[1]);
		if (getchar() != 'y') {
			printf("Exiting...\n");
			exit(-1);
		}
		
		/*TODO - Test the existing file or device for adequate space*/
		fd = open(argv[1], O_RDWR | O_LARGEFILE);
		if (fd < 0) {
			printf("ERROR: opening file [%s] (errno %d)\n",
			       argv[1], 0 - errno);
			exit(-1);
		}

		if (S_ISBLK(finfo.st_mode)) {
			if(ioctl(fd,BLKGETSIZE,&size)!=0) {
				printf("ERROR: BLKGETSIZE failed, "
					"couldn't stat image [%s]\n", 
					argv[1]);
				close(fd);
				exit(-1);
			}
			if (size < ddqcow.td_state->size<<9) {
				printf("ERROR: Not enough space on device "
					"%s (%lu bytes available, %llu bytes required\n",
					argv[1], size, 
					(long long unsigned)ddqcow.td_state->size<<9);
				close(fd);
				exit(-1);				
			}
		} else {
			if (ftruncate(fd, (off_t)ddqcow.td_state->size<<9) < 0) {
				printf("Unable to create file "
					"[%s] of size %llu (errno %d). "
					 "Exiting...\n",
					argv[1], 
					(long long unsigned)ddqcow.td_state->size<<9, 
					 0 - errno);
				close(fd);
				exit(-1);
			} else printf("File [%s] truncated to length %llu "
					"(%llu)\n", 
				       argv[1], 
				       (long long unsigned)ddqcow.td_state->size<<9, 
				       (long long unsigned)ddqcow.td_state->size);
		}
		close(fd);
	}

	/* Open destination file */
	if ((aio_fd = open(argv[1], O_RDWR | O_LARGEFILE)) == -1) {
		printf("Can't open target image %s\n", argv[1]);
		exit(-1);
	}

	/*Initialise the output string*/
	memset(output,0x20,25);
	output[0] = '[';
	output[22] = ']';
	output[23] = '\0';
	printf("%s",output);

	i = 0;
	while (running) {
		int wait = 0;

		if (!complete) {
			/*Read Pages from qcow image*/
			if ( (ret = posix_memalign((void **)&buf, 
						   BLOCK_PROCESSSZ, 
						   BLOCK_PROCESSSZ))
			     != 0) {
				printf("Unable to alloc memory (%d)\n",ret);
				exit(-1);				
			}
		
			/*Attempt to read 4k sized blocks*/
			ret = ddqcow.drv->td_queue_read(&ddqcow, i>>9,
							BLOCK_PROCESSSZ>>9, buf, 
							send_read_responses, i>>9, buf);

			if (ret < 0) {
				if (ret == -EBUSY)
					wait = 1;
				else {
					printf("UNABLE TO READ block [%llu]\n",
					       (long long unsigned)i);
					exit(-1);
				}
			} else {
				i += BLOCK_PROCESSSZ;
				submit_events++;
			}

			if (i >= ddqcow.td_state->size<<9) {
				complete = 1;
			}
			
			if ((submit_events % 10 == 0) || complete || wait) 
				ddqcow.drv->td_submit(&ddqcow);
		}
		
		if (complete) {
			if (returned_write_events == submit_events)
				running = wait = 0;
			else
				wait = 1;
		}

		/*Check AIO FD*/
		LOCAL_FD_SET(&readfds);
                ret = select(maxfds + 1, &readfds, NULL, NULL,
                             wait ? NULL :  &timeout);
			     
		if (ret > 0) {
			if (ddqcow.drv->td_do_callbacks(&ddqcow, 0) < 0) {
				printf("Error on reading callback\n");
				exit(-1);
			}
		}
	}
	close(aio_fd);
	memcpy(output+prev+1,"=",1);
	printf("\r%s     100%%\nTRANSFER COMPLETE\n\n", output);
		
	return 0;
}
